import asyncio
import enum
import os, torch
import json
import queue
import threading
import uuid
from typing import Any, List, Tuple
from aiges.core.types import *

try:
    from aiges_embed import ResponseData, Response, DataListNode, DataListCls, SessionCreateResponse, callback  # c++
except:
    from aiges.dto import Response, ResponseData, DataListNode, DataListCls, SessionCreateResponse, callback

from aiges.sdk import WrapperBase
from aiges.utils.log import log, getFileLogger
from transformers import BertTokenizer, AutoTokenizer
import logging
import threading
import numpy as np
import onnxruntime
from ailab.inference_wrapper.huggingface.transformers.bert import transform_list_of_texts


DataNone = -1
DataBegin = 0  # 首数据
DataContinue = 1  # 中间数据
DataEnd = 2  # 尾数据

class RequestMode(enum.Enum):
    ONCE = 'once'
    ONCE_ASYNC = 'once_async'
    STREAM = 'stream'
    STOP = "stop"

class RequestInfo:
    def __init__(self, sid: str, params: dict, user_tag: str="") -> None:
        self.handle = str(uuid.uuid4())
        self.sid = sid
        self.params = params
        self.user_tag = user_tag
        self.request = []      

class PromptInferenceInfo:
    def __init__(self, wrapper,
                 thread_id: str,
                 mode: RequestMode,
                 prompt: str,
                 requestInfo: RequestInfo,
                 result_q: queue.Queue = None):
        self.wrapper = wrapper
        self.thread_id = thread_id
        self.mode = mode
        self.prompt = prompt
        self.requestInfo = requestInfo
        self.result_q = result_q
        self.request_id = str(uuid.uuid4().hex)


elapsed_times = 0
import functools
import time
def time_cost(func):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = await func(*args, **kwargs)
        end_time = time.perf_counter()
        global elapsed_times
        elapsed_time = end_time - start_time
        elapsed_times += elapsed_time
        print(f"函数 {func.__name__} 的耗时为: {elapsed_time:.4f} 秒")
        return result
    return wrapper

def get_elapsed_time():
    return elapsed_times

def resp_content(status, index, text):
    resd = ResponseData()
    resd.key = "content"
    resd.setDataType(DataText)
    resd.status = status
    resd.setData(json.dumps(text).encode('utf-8'))
    return resd

def resp_usage(status, prompt_token, completion_token):
    resp_json = {"completion_tokens": completion_token, "prompt_tokens": prompt_token,
                 "total_tokens": prompt_token + completion_token}
    resd = ResponseData()
    resd.key = "usage"
    resd.setDataType(DataText)
    resd.status = status
    resd.setData(json.dumps(resp_json).encode('utf-8'))
    return resd

def get_payload_messages(reqData: DataListCls):
    messagesStr = reqData.get('messages').data.decode('utf-8')
    # 校验格式
    try:
        messages = json.loads(messagesStr)
        message = ""
        for item in messages['messages']:
            if item['role'] == "user":
                message = item['content']
    except Exception as e:
        import traceback
        traceback.print_exc()
        return None
    # self.filelogger.info(f'message: {message}')
    batch = []
    batch.append(message)
    return batch
    
class ThreadPool:
    def __init__(self, num_threads, wrapper):
        self.num_threads = num_threads
        self.threads = {}
        self.task_queues = {}
        self.lock = threading.Lock()
        self.wrapper = wrapper
        self.max_batch_size = 8
        self.filelogger = getFileLogger(level=logging.INFO)

        for i in range(num_threads):
            process_id = os.getpid
            thread_id = f"process-{str(process_id)}-thread-{str(i)}"
            task_queue = queue.Queue()

            self.threads[thread_id] = threading.Thread(target=self.worker, args=(thread_id, task_queue, self.max_batch_size))
            self.task_queues[thread_id] = task_queue

        for thread in self.threads.values():
            thread.start()  


    async def bert_loop(self, task_queue, max_batch_size):
        self.filelogger.info(f"bert_loop started")
        tasks = []
        while True:
            await asyncio.sleep(0.01)

            while not task_queue.empty() and len(tasks) < max_batch_size:
                task: PromptInferenceInfo = task_queue.get_nowait()
                if task is None:
                    break
                tasks.append(task)

            if tasks:
                self.filelogger.info(f"Processing {len(tasks)} tasks.")
                asyncio.create_task(task.wrapper.infer(tasks))
                tasks = [] 


    def worker(self, thread_id, task_queue, max_batch_size):
        # 创建事件循环
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(self.bert_loop(task_queue, max_batch_size))

    def alloc_min_thread(self) -> str:
        with self.lock:
            min_thread_id = min(self.threads, key=lambda thread_id: self.task_queues[thread_id].qsize())
        return min_thread_id
    
    def put_task(self, thread_id, task):
        with self.lock:
            self.task_queues[thread_id].put(task)

    def wait_completion(self):
        for task_queue in self.task_queues.values():
            task_queue.put(None)
        for thread in self.threads.values():
            thread.join()


class Wrapper(WrapperBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filelogger = getFileLogger(level=logging.INFO)
        self.patch_id = {}
        self.lock = threading.Lock()
        self.request_map = {}
        self.request_map_lock = threading.Lock()
        self.thread_pool_size = 1
        self.thread_pool = None
        
    def wrapperInit(self, config: dict) -> int:
        self.filelogger.info("Initializing ...")
        self.pretrained_model = os.environ.get("PRETRAINED_MODEL_NAME")
        self.base_model_path = os.environ.get("FULL_MODEL_PATH")
        self.tokenizer = AutoTokenizer.from_pretrained(self.base_model_path, trust_remote_code=True)
        self.thread_pool = ThreadPool(num_threads=self.thread_pool_size, wrapper=self)
        id2label_file = os.path.join(self.base_model_path, 'id2label.json')
        with open(id2label_file, 'r') as file:
            self.tag_data = json.load(file)
        self.cls_mode = 'sigmoid'
        self.pooling_strategy = 'max'
        self.filelogger.info(f'wrapper init success, create thread:{self.thread_pool_size}')
        return 0
    
    def wrapperFini(self):
        self.thread_pool.wait_completion()
        return 0
    
    def wrapperCreate(self, params: dict, sid: str, persId: int=0, usrTag: str="") -> SessionCreateResponse:
        patch_id = str(params.get("patch_id", "0"))
        self.filelogger.info(f'start wrapperCreate {params}')
        if len(patch_id) == 0:
            patch_id = "0"
        requestInfo = RequestInfo(sid, params, usrTag)
        self.request_map_lock.acquire()
        self.request_map[requestInfo.handle] = requestInfo
        self.request_map_lock.release()

        s = SessionCreateResponse()
        s.handle = requestInfo.handle
        s.error_code = 0
        self.filelogger.debug(f'success wrapperCreate {patch_id}, handle {requestInfo.handle}')
        return s

    def wrapperWrite(self, handle: str, req: DataListCls) -> str:
        self.filelogger.info(f'start wrapperWrite handle {handle}')
        prompt = get_payload_messages(req)
        if prompt is None:
            self.filelogger.error("parse prompt failed")
            return -1
        self.request_map_lock.acquire()
        requestInfo = self.request_map[handle]
        if requestInfo == None:
            self.filelogger.error(f"can't get this handle: {handle}")
            self.request_map_lock.release()
            return -1
        self.request_map_lock.release()
        thread_id = self.thread_pool.alloc_min_thread()
        inferenceInfo = PromptInferenceInfo(self, thread_id, RequestMode.STREAM, prompt, requestInfo)
        self.thread_pool.put_task(thread_id, inferenceInfo)
        self.request_map_lock.acquire()
        self.request_map[handle].request.append(inferenceInfo.request_id)
        self.request_map_lock.release()
        self.filelogger.debug(
            f'success wrapperWrite handle:{handle}, thread_id:{thread_id}, request_id:{inferenceInfo.request_id}')
        return 0

    def wrapperDestroy(self, handle: str):
        self.filelogger.debug(f'start wrapperDestroy {handle}')
        self.request_map_lock.acquire()
        requestInfo = self.request_map[handle]
        if requestInfo == None:
            self.filelogger.error(f"can't get this handle: {handle}")
            self.request_map_lock.release()
            return -1
        self.request_map_lock.release()
        self.filelogger.debug(f'success wrapperDestroy {handle}')
        return 0

    def wrapperError(self):
        pass

    def wrapperOnceExec(self):
        pass

    def wrapperOnceExecAsync(self, params: list, reqData: DataListCls, sid: str, usrTag: str, persId: int=0):
        prompt = get_payload_messages(reqData)
        if prompt == None:
            self.filelogger.error(f'parse prompt failed')
            return -1
        requestInfo = RequestInfo("", params, usrTag)
        thread_id = self.thread_pool.alloc_min_thread()
        result_q = queue.Queue()
        inferenceInfo = PromptInferenceInfo(self, thread_id, RequestMode.ONCE_ASYNC, prompt, requestInfo, result_q)
        self.filelogger.info(f'start wrapperOnceExecAsync params:{params}, request_id:{inferenceInfo.request_id}')
        self.thread_pool.put_task(thread_id, inferenceInfo)
        
        self.filelogger.debug(
            f'success wrapperOnceExecAsync, params:{params}, prompt:{prompt}, request_id:{inferenceInfo.request_id}')
        return 0

    async def infer(self, tasks: list):
        task_data = {
            "request_ids": {},
            "request_infos": {},
            "params": {},
            "patch_ids": {},
            "user_tags": {},
            "prompts": []
        }
        for idx, inferenceInfo in enumerate(tasks):
            task_data["request_ids"][idx] = inferenceInfo.request_id
            task_data["request_infos"][idx] = inferenceInfo.requestInfo
            task_data["params"][idx] = task_data["request_infos"][idx].params
            task_data["user_tags"][idx] = task_data["request_infos"][idx].user_tag
            task_data["patch_ids"][idx] = str(task_data["params"][idx].get("patch_id", 0))
            task_data["prompts"].extend(inferenceInfo.prompt)  
        self.filelogger.info(f'prompts: {task_data["prompts"]}')

        onnx_tokenized = transform_list_of_texts.transform_list_of_texts(
            task_data["prompts"], tokenizer=self.tokenizer, 
            chunk_size=510, stride=510, minimal_chunk_length=510, maximal_text_length=2040
        )
        input_tokens = self.get_tokens(onnx_tokenized)
        self.filelogger.info(f'input_tokens: {input_tokens}')
        # self.filelogger.info(f'onnx_tokenized: {onnx_tokenized}, len: {onnx_tokenized["attention_mask"][0].sum()}')

        sess_options = onnxruntime.SessionOptions()
        sess_options.intra_op_num_threads = 5
        sess_options.inter_op_num_threads = 1
        sess_options.execution_mode = onnxruntime.ExecutionMode.ORT_SEQUENTIAL
        sess_options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_EXTENDED        
        onnx_model_path = os.path.join(self.base_model_path, 'model.onnx')
        providers = ['CUDAExecutionProvider', 'CPUExecutionProvider']
        session = onnxruntime.InferenceSession(onnx_model_path, sess_options=sess_options, providers=providers)
        nums_question = len(task_data["prompts"])
        res_tags = await self.onnx_predict(onnx_tokenized, session, nums_question)
        user_tags = task_data["user_tags"]
        for i, res_tag in enumerate(res_tags):
            res = Response()
            content = resp_content(Once, 0, res_tag)
            usage = resp_usage(Once, int(input_tokens[i]), 0)
            res.list = [content, usage]
            callback(res, user_tags[i])
            self.filelogger.info(f'====> inference end, user_tag: {user_tags[i]}, res: {res.list[0].data}, prompt_tokens: {input_tokens[i]}')
        return

    # .onnx 格式模型推理         
    @time_cost       
    async def onnx_predict(self, onnx_tokenized, session, nums_question):
        res_list = []
        article_chunk_nums = []
        batch_size = nums_question if nums_question <= self.thread_pool.max_batch_size else self.thread_pool.max_batch_size
        max_len = 512
        input_ids = np.empty((0, max_len))
        attention_mask = np.empty((0, max_len))
        for idx in range(batch_size):
            input_ids_single = onnx_tokenized["input_ids"][idx].astype('int64')
            attention_mask_single = onnx_tokenized["attention_mask"][idx].astype('float32')
            input_ids = np.concatenate((input_ids, input_ids_single), axis=0)
            attention_mask = np.concatenate((attention_mask, attention_mask_single), axis=0)
            article_chunk_nums.append(len(input_ids_single))

        ort_inputs = {
            'input_ids': input_ids.astype('int64'),
            'attention_mask': attention_mask.astype('float32'),
        }
        ort_outputs = session.run(None, ort_inputs)[0]
        # self.filelogger.info(f'\nort_outputs: {ort_outputs}')

        # 归一化处理
        split_indices = np.array(article_chunk_nums).cumsum()[:-1]
        logits_split = np.split(ort_outputs, split_indices)
        logits_output = []
        for i, logit in enumerate(logits_split):
            if self.pooling_strategy == "mean":
                pooled_logits = np.mean(logit, axis=0)
            elif self.pooling_strategy == "max":
                pooled_logits = np.max(logit, axis=0)
            else:
                raise ValueError("Unknown pooling strategy!")
            logits_output.append(pooled_logits)

        def softmax(logits):
            exp_logits = np.exp(logits - np.max(logits, axis=-1, keepdims=True)) 
            return exp_logits / np.sum(exp_logits, axis=-1, keepdims=True)
        
        def sigmoid(logits):
            return 1 / (1 + np.exp(-logits))

        # 将 logits_output 转换为概率
        if self.cls_mode == 'softmax':
            probabilities = softmax(np.array(logits_output))
        elif self.cls_mode == 'sigmoid':
            probabilities = sigmoid(np.array(logits_output))

        res_list = []
        for i, p in enumerate(probabilities):
            self.filelogger.info(f"text_i: {i}")
            category_scores = {}
            for idx, prob in enumerate(p):
                # self.filelogger.info(f'idx: {idx}, prob: {prob}')
                category = self.tag_data.get(str(idx), None)
                if category:
                    category_scores[category] = float(prob)
            res_tag = {
                "category_scores": category_scores
            }
            res_list.append(res_tag)
        self.filelogger.info(f'res_list: {res_list}')
        return res_list  
    
    def get_tokens(self, onnx_tokenized):
        tokens_list = []
        for idx in range(len(onnx_tokenized["input_ids"])):
            tokens_list.append(onnx_tokenized['attention_mask'][idx].sum())
        return tokens_list
